#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Copyright (C) 2018  Canonical Ltd.
# Author: Mathieu Trudel-Lapierre <cyphermox@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Analyze britney's excuses output and suggest a course of action for
proposed migration.
"""

# FIXME: Refactor this majorly
#
# FIXME: Various parts of slangasek's pseudocode (in comments where relevant)
#        are not well implemented.
#
# TODO: finish moving things from requests to urllib; no sense in using two
#       different modules for URL retrieval.

import yaml
import os
import time
import subprocess
import argparse
import requests
import tempfile
import logging

from enum import Enum
from urllib.request import FancyURLopener
from launchpadlib.launchpad import Launchpad

DEBIAN_CURRENT_SERIES = 'sid'
ARCHIVE_PAGES = 'https://people.canonical.com/~ubuntu-archive/'
MAX_CACHE_AGE = 14400   # excuses cache should not be older than 4 hours

excuses = {}

level = 0
seen = []

assistant_logger = logging.getLogger(__name__)


class ExcuseValue(Enum):

    NONE = 1
    PASS = 2
    FAIL = 3
    INFO = 4


class StatusAdapter(logging.LoggerAdapter):

    depth = 0

    def process(self, msg, kwargs):
        status = kwargs.pop('status')

        # FIXME: identationing may be ugly because of character width
        if status is ExcuseValue.PASS:
            icon = "\033[92m✔\033[0m"
        elif status is ExcuseValue.FAIL:
            icon = "\033[91m✘\033[0m"
        elif status is ExcuseValue.INFO:
            icon = "\033[94m\033[0m"
        else:
            icon = ""

        if self.depth <= 0:
            return '%s %s' % (msg, icon), kwargs
        elif status is ExcuseValue.INFO:
            return '%s- %s %s' % (" " * self.depth * 2, icon, msg), kwargs
        else:
            return '%s- %s %s' % (" " * self.depth * 2, msg, icon), kwargs

    def critical(self, msg, *args, **kwargs):
        self.depth = self.extra['depth']
        msg, kwargs = self.process(msg, kwargs)
        self.logger.critical(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.depth = self.extra['depth'] + 1
        msg, kwargs = self.process(msg, kwargs)
        self.logger.error(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.depth = self.extra['depth'] + 2
        msg, kwargs = self.process(msg, kwargs)
        self.logger.warning(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.depth = self.extra['depth'] + 3
        msg, kwargs = self.process(msg, kwargs)
        self.logger.info(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.depth = self.extra['depth']
        msg, kwargs = self.process("DEBUG: {}".format(msg), kwargs)
        self.logger.debug(msg, *args, **kwargs)


def report_download(blocks_read, block_size, total_size):
    size_read = blocks_read * block_size
    percent = size_read/total_size*100
    print("Refreshing excuses: %.0f %%" % (percent), end='\r')


def get_debian_ci_results(source, arch):
    try:
        url = "https://ci.debian.net/data/packages/unstable/{}/{}/latest.json"
        results_url = url.format("amd64", get_pkg_archive_path(source))
        resp = requests.get(url=results_url)
        return resp.json()
    except Exception:
        return None


def find_excuses(lp, src, level, seen):
    if src in seen:
        return

    for item in excuses['sources']:
        item_name = item.get('item-name')

        if item_name == src:
            process(lp, item, level, seen)


def get_pkg_archive_path(package):
    try:
        # TODO: refactor to avoid shell=True
        path = subprocess.check_output(
            "apt-cache show %s | grep Filename:" % package,
            shell=True, universal_newlines=True)
        path = path.split(' ')[1].split('/')
        path = "/".join(path[2:4])
        return path
    except Exception:
        return None


def get_source_package(binary):
    source = None
    # TODO: refactor to avoid shell=True
    try:
        source = subprocess.check_output(
            "apt-cache show %s | grep Source:" % binary,
            shell=True, universal_newlines=True)
    except subprocess.CalledProcessError as e:
        source = subprocess.check_output(
            "apt-cache show %s | grep Package:" % binary,
            shell=True, universal_newlines=True)

    if source is not None:
        if source.startswith("Source:") or source.startswith("Package:"):
            name = source.split()[1]
            return name

    return None


def package_in_distro(package, distro='ubuntu', distroseries='bionic',
                      proposed=False):
    assistant = StatusAdapter(assistant_logger, {'depth': 0})

    # TODO: This operation is pretty costly, do caching?

    if distro == 'debian':
        distroseries = DEBIAN_CURRENT_SERIES
    if proposed:
        distroseries += "-proposed"

    madison_url = "https://qa.debian.org/cgi-bin/madison.cgi"
    params = "?package={}&table={}&a=&c=&s={}".format(package,
                                                      distro,
                                                      distroseries)
    url = madison_url + params
    resp = requests.get(url=url)

    package_found = {}
    for line in resp.text.split('\n'):
        if " {} ".format(package) not in line:
            continue
        package_line = line.split(' | ')

        series_component = package_line[2].split('/')
        component = 'main'
        if len(series_component) > 1:
            component = series_component[1]

        if '{}'.format(distroseries) in series_component[0]:
            if distro == 'ubuntu':
                package_found = {
                    'version': package_line[1],
                    'component': component,
                }
            else:
                package_found = {
                    'version': package_line[1],
                }

            return package_found

    return {}


def process_lp_build_results(source, archive, series, level, uploads, failed):
    assistant = StatusAdapter(assistant_logger, {'depth': level + 1})

    source_name = source.get('source')

    spph = archive.getPublishedSources(exact_match=True,
                                       source_name=source_name,
                                       distro_series=series,
                                       pocket="Proposed",
                                       order_by_date=True)

    new_version = series.getPackageUploads(archive=archive,
                                           name=source_name,
                                           version=source.get('new-version'),
                                           pocket="Proposed",
                                           exact_match=True)

    for item in new_version:
        arch = item.display_arches.split(',')[0]
        if item.package_version not in uploads:
            uploads[item.package_version] = {}
        if arch == 'source':
            continue
        uploads[item.package_version][arch] = item.getBinaryProperties()

    # Only get the builds for the latest publication, this is more likely to
    # be new source in -proposed, or the most recent upload.
    builds = spph[0].getBuilds()
    for build in builds:
        if "Successfully" not in build.buildstate:
            assistant.error("{} is missing a build on {}:".format(
                                source_name, build.arch_tag),
                            status=ExcuseValue.FAIL)
            log_url = build.build_log_url
            if not log_url:
                log_url = "<No build log available>"
            assistant.warning("[%s] %s" % (build.buildstate,
                                           log_url),
                              status=ExcuseValue.NONE)
            failed[build.arch_tag] = {
                'state': build.buildstate,
            }


def process_unsatisfiable_depends(source, series, level, seen):
    assistant = StatusAdapter(assistant_logger, {'depth': level + 1})

    affected_sources = {}
    unsatisfiable = {}
    distroseries = series.name

    depends = source.get('dependencies').get('unsatisfiable-dependencies')
    for arch, signatures in depends.items():
        unsatisfiable[arch] = set()
        affected_sources[arch] = set()
        for signature in signatures:
            unsatisfiable[arch].add(signature)
            binary_name = signature.split(' ')[0]
            try:
                pkg = get_source_package(binary_name)
                affected_sources[arch].add(pkg)
            except Exception:
                # FIXME: we might be dealing with a new package in proposed
                #        here, but using the binary name instead of the source
                #        name.
                if any(package_in_distro(binary_name, distro='ubuntu',
                                         distroseries=distroseries)):
                    affected_sources[arch].add(binary_name)
                elif any(package_in_distro(binary_name,
                                           distro='ubuntu',
                                           distroseries=distroseries,
                                           proposed=True)):
                    affected_sources[arch].append(binary_name)

    if len(affected_sources.items()) < 1 and len(unsatisfiable.items()) < 1:
        return

    assistant.critical("Fix unsatisfiable dependencies in {}:".format(
                           source.get('source')),
                       status=ExcuseValue.NONE)

    # TODO: Check for MIR bugs
    #  - is there an MIR?
    #    - has the MIR been rejected?
    #      - upload or submit to sponsorship queue to drop the dependency
    #  - else
    #    - file an MIR

    # TODO: Check version comparisons for removal requests/fixes
    # - is the unsatisfied dependency due to a package dropped in Ubuntu,
    #   but not in Debian, which may come back as a sync later
    #   (i.e. not blacklisted)?
    #   - leave in -proposed
    # - is this package Ubuntu-specific?
    #   - is there an open bug in launchpad about this issue, with no action?
    #     - subscribe ubuntu-archive and request the package's removal
    #   - else
    #     - open a bug report and assign to the package's maintainer
    # - is the package in Debian, but the dependency is part of Ubuntu delta?
    #   - fix

    for arch, signatures in unsatisfiable.items():
        for signature in list(signatures):
            depends = signature.split(' ')[0]
            assistant.error("{} can not be satisfied".format(signature),
                            status=ExcuseValue.FAIL)
            in_archive = package_in_distro(depends, distro='ubuntu',
                                           distroseries=distroseries)
            in_proposed = package_in_distro(depends, distro='ubuntu',
                                            distroseries=distroseries,
                                            proposed=True)

            if any(in_archive) and not any(in_proposed):
                assistant.warning("{}/{} exists "
                                  "in the Ubuntu primary archive".format(
                                      depends,
                                      in_archive.get('version')),
                                  status=ExcuseValue.FAIL)
                if source.get('component', 'main') != in_archive.get('component'):
                    assistant.info("This may be a component mismatch",
                                   status=ExcuseValue.INFO)
                    assistant.info("Check if a MIR bug exists, if not, create one",
                                   status=ExcuseValue.INFO)
            elif not any(in_archive) and any(in_proposed):
                assistant.warning("{} is only in -proposed".format(depends),
                                  status=ExcuseValue.FAIL)
                assistant.info("Has this package been dropped in Ubuntu, "
                               "but not in Debian?",
                               status=ExcuseValue.INFO)
            elif not any(in_archive) and not any(in_proposed):
                in_debian = package_in_distro(depends, distro='debian',
                                              distroseries=distroseries)
                if any(in_debian):
                    assistant.warning("{} only exists in Debian".format(depends),
                                      status=ExcuseValue.FAIL)
                    assistant.info("Is this package blacklisted? Should it be synced?",
                                   status=ExcuseValue.INFO)
                else:
                    assistant.warning("{} is not found".format(depends),
                                      status=ExcuseValue.FAIL)
                    assistant.info("Has this package been removed?",
                                   status=ExcuseValue.INFO)
                    assistant.info("Consider filing a bug for the removal of "
                                   "{}".format(source.get('source')),
                                   status=ExcuseValue.INFO)
            else:
                if source.get('component', 'main') != in_archive.get('component'):
                    assistant.info("This may be a component mismatch",
                                   status=ExcuseValue.INFO)
                    assistant.info("Check if a MIR bug exists, if not, create one",
                                   status=ExcuseValue.INFO)

    if len(affected_sources) > 0:
        for src in affected_sources:
            find_excuses(src, level+2, seen)


def process_autopkgtest(source, level):
    assistant = StatusAdapter(assistant_logger, {'depth': level + 1})

    autopkgtests = source.get('policy_info').get('autopkgtest')

    assistant.critical("Fix autopkgtests triggered by this package for:",
                       status=ExcuseValue.NONE)

    waiting = 0
    for key, test in autopkgtests.items():
        for arch, arch_test in test.items():
            if 'RUNNING' in arch_test:
                waiting += 1
            if 'REGRESSION' in arch_test:
                assistant.error("{} {} {}".format(key, arch, arch_test[2]),
                                status=ExcuseValue.FAIL)
                if arch == "amd64":
                    if '/' in key:
                        pkgname = key.split('/')[0]
                    else:
                        pkgname = key
                    ci_results = get_debian_ci_results(pkgname, "amd64")
                    if ci_results is not None:
                        result = ci_results.get('status')
                        status_ci = ExcuseValue.FAIL
                        if result == 'pass':
                            status_ci = ExcuseValue.PASS
                        assistant.warning("CI tests {} in Debian".format(
                                                result),
                                          status=status_ci)
                        if 'pass' in result:
                            assistant.info("Consider filing a bug "
                                           "(usertag: autopkgtest) "
                                           "in Debian if none exist",
                                           status=ExcuseValue.INFO)
                        else:
                            # TODO: (cyphermox) detect this case?
                            #       check versions?
                            assistant.info("If synced from Debian and "
                                           "requires sourceful changes to "
                                           "the package, file a bug for "
                                           "removal from -proposed",
                                           status=ExcuseValue.INFO)

    if waiting > 0:
        assistant.error("{} tests are currently running "
                        "or waiting to be run".format(waiting),
                        status=ExcuseValue.INFO)


def process_blocking(source, lp, level):
    assistant = StatusAdapter(assistant_logger, {'depth': level + 1})

    bugs = source.get('policy_info').get('block-bugs')

    assistant.critical("Resolve blocking bugs:",
                       status=ExcuseValue.NONE)

    for bug in bugs.keys():
        lp_bug = lp.bugs[bug]
        assistant.error("[LP: #{}] {} {}".format(lp_bug.id,
                                                 lp_bug.title,
                                                 lp_bug.web_link),
                        status=ExcuseValue.NONE)
        tasks = lp_bug.bug_tasks
        for task in tasks:
            value = ExcuseValue.FAIL
            if task.status in ('Fix Committed', 'Fix Released'):
                value = ExcuseValue.PASS
            elif task.status in ("Won't Fix", 'Invalid'):
                continue
            assistant.warning("{}({}) in {}".format(
                                     task.status,
                                     task.importance,
                                     task.bug_target_display_name),
                              status=value)

        # guesstimate whether this is a removal request
        if 'emove {}'.format(source.get('source')) in lp_bug.title:
            assistant.info("This looks like a removal request",
                           status=ExcuseValue.INFO)
            assistant.info("Consider pinging #ubuntu-release for processing",
                           status=ExcuseValue.INFO)

    hints = source.get('hints')
    if hints is not None:
        assistant.critical("Update manual hinting (contact #ubuntu-release):",
                           status=ExcuseValue.NONE)
        # TODO: fix hint output.
        # Unblock request by apw ignored due to version mismatch: 4.4.0-1081.86
        # Not touching package due to block request by apw (contact #ubuntu-release if update is needed)
        #  -- missing in excuses.yaml; only the hint-from and hint-type are present.
        #
        # for excuse in source.get('excuses'):
        #    if 'block request' not in excuse:
        #        continue
        #    assistant.error(excuse, status=ExcuseValue.INFO)


def process_dependencies(source, level, seen):
    assistant = StatusAdapter(assistant_logger, {'depth': level + 1})

    dependencies = source.get('dependencies')
    blocked_by = dependencies.get('blocked-by', None)
    migrate_after = dependencies.get('migrate-after', None)

    assistant.critical("Clear outstanding promotion interdependencies:",
                       status=ExcuseValue.NONE)

    if migrate_after is not None:
        assistant.error("{} will migrate after {}".format(
                           source.get('source'), ", ".join(migrate_after)),
                        status=ExcuseValue.FAIL)
        assistant.warning("Investigate what packages are conflicting, "
                          "by looking at 'Trying easy on autohinter' lines in "
                          "update_output.txt for {}".format(
                                source.get('source')),
                          status=ExcuseValue.INFO)
        assistant.warning("See {}proposed-migration/update_output.txt".format(
                              ARCHIVE_PAGES),
                          status=ExcuseValue.INFO)

    if blocked_by is not None:
        assistant.error("{} is blocked by the migration of {}".format(
                           source.get('source'), ", ".join(blocked_by)),
                        status=ExcuseValue.FAIL)
        for blocker in blocked_by:
            find_excuses(blocker, level+2, seen)


def process_missing_builds(source, archive, series, level):
    assistant = StatusAdapter(assistant_logger, {'depth': level + 1})

    source_name = source.get('source')
    new_version = source.get('new-version')
    old_version = source.get('old-version')
    old_binaries = source.get('old-binaries').get(old_version)
    reasons = source.get('reason')

    # TODO: Process missing builds; suggest options
    #
    #  - missing build on $arch / has no binaries on any arch
    #    - is this an architecture-specific build failure?
    #    - has Debian removed the binaries for this architecture?
    #    - ask AA to remove the binaries as ANAIS
    #  - else
    #    - try to fix
    #
    #  - is this a build failure on all archs?
    #    - are there bugs filed about this failure in Debian?
    #      - is the package in sync with Debian and does the package require
    #        sourceful changes to fix?
    #        - remove from -proposed
    #
    #  - does the package fail to build in Debian?
    #    - file a bug in Debian
    #    - is the package in sync with Debian and does the package require
    #      sourceful changes to fix?
    #      - remove from -proposed
    #
    #  - is this a dep-wait?
    #    - does this package have this build-dependency in Debian?
    #      - is this an architecture-specific dep-wait?
    #        - has Debian removed the binaries for this architecture?
    #          - ask AA to remove the binaries as ANAIS
    #        - else
    #          - try to fix
    #      - does this binary package exist in Debian?
    #        - look what source package provides this binary package in Debian
    #        - is this source package ftbfs or dep-wait in -proposed?
    #          - recurse
    #        - else
    #          - is this source package on the sync blacklist?
    #            - file a bug with the Ubuntu package
    #          - else
    #            - fix by syncing or merging the source
    #      - else
    #        - make sure a bug is filed in Debian about the issue
    #        - was the depended-on package removed from Debian,
    #          and is this a sync?
    #          - ask AA to remove the package from -proposed
    #        - else
    #          - leave the package in -proposed

    assistant.critical("Fix missing builds:", status=ExcuseValue.NONE)

    uploads = {}
    failed = {}
    new = []
    nbs = set()
    anais = []
    new_binaries = set()

    process_lp_build_results(source, archive, series, level, uploads, failed)

    if new_version in uploads:
        for arch, item in uploads[new_version].items():
            for binary in item:
                binary_name = binary.get('name')
                new_binaries.add(binary_name)
                if binary.get('is_new'):
                    new.append(binary)

    if not any(failed):
        assistant.error("No failed builds found", status=ExcuseValue.PASS)

        try:
            missing_builds = source.get('missing-builds')
            missing_arches = missing_builds.get('on-architectures')
            arch_o = []
            for arch in missing_arches:
                if arch not in uploads[new_version]:
                    arch_o.append("-a {}".format(arch))

            if any(arch_o):
                assistant.warning("This package has dropped support for "
                                  "architectures it previous supported. ",
                                  status=ExcuseValue.INFO)
                assistant.warning("Ask in #ubuntu-release for an Archive "
                                  "Admin to run:",
                                  status=ExcuseValue.INFO)
                assistant.info("remove-package %(arches)s -b %(bins)s"
                               % ({'arches': " ".join(arch_o),
                                   'bins': " ".join(old_binaries),
                                   }), status=ExcuseValue.NONE)
        except AttributeError:
            # Ignore a failure here, it just means we don't have
            # missing-builds to process after all.
            pass

    if any(new):
        assistant.error("This package has NEW binaries to process:",
                        status=ExcuseValue.INFO)
        for binary in new:
            assistant.warning("[{}] {}/{}".format(
                                  binary.get('architecture'),
                                  binary.get('name'),
                                  binary.get('version')),
                              status=ExcuseValue.FAIL)


def process(lp, source, level, seen):
    assistant = StatusAdapter(assistant_logger, {'depth': level})

    ubuntu = lp.distributions["ubuntu"]
    archive = ubuntu.main_archive
    series = ubuntu.current_series

    source_name = source.get('source')
    reasons = source.get('reason')

    seen.append(source_name)

    text_candidate = "not considered"
    candidate = ExcuseValue.FAIL
    if source.get('is-candidate'):
        text_candidate = "a valid candidate"
        candidate = ExcuseValue.PASS
    assistant.critical("{} is {}".format(source_name, text_candidate),
                       status=candidate)

    assistant.critical("Next steps for {} {}: {}".format(
                            source_name, source.get('new-version'), reasons),
                       status=ExcuseValue.NONE)

    work_needed = False

    missing_builds = source.get('missing-builds')
    if missing_builds is not None or 'no-binaries' in reasons:
        work_needed = True
        process_missing_builds(source, archive, series, level)

    if 'depends' in reasons:
        work_needed = True
        process_unsatisfiable_depends(source, series, level, seen)

    if 'block' in reasons:
        work_needed = True
        process_blocking(source, lp, level)

    if 'autopkgtest' in reasons:
        work_needed = True
        process_autopkgtest(source, level)

    dependencies = source.get('dependencies')
    if dependencies is not None:
        work_needed = True
        process_dependencies(source, level, seen)

    if work_needed is False:
        assistant.error("Good job!", status=ExcuseValue.PASS)
        assistant.warning("Investigate if packages are conflicting, "
                          "by looking at 'Trying easy on autohinter' lines in "
                          "update_output.txt"
                          " for {}".format(source.get('source')),
                          status=ExcuseValue.INFO)
        assistant.warning("See {}proposed-migration/update_output.txt".format(
                               ARCHIVE_PAGES),
                          status=ExcuseValue.INFO)


def choose_blocked_source(excuses):
    import textwrap

    wrapper = textwrap.TextWrapper(break_long_words=False, break_on_hyphens=False)

    choice = 0
    options = {}
    src_num = 1
    msg = ""
    for item in excuses['sources']:
        item_name = item.get('item-name')
        options[src_num] = item_name
        msg += "({}){} ".format(src_num, item_name)
        src_num += 1

    while True:
        print(options)
        print("\n".join(wrapper.wrap(msg)))
        num = input("\nWhich package do you want to look at?")

        try:
            choice = int(num)
            if choice > 0 and choice < src_num:
                break
        except ValueError as e:
            # num might be the package name.
            if num in options.values():
                return num

    return options[choice]


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description='Evaluate next steps for proposed migration')
    parser.add_argument('-s', '--source', dest='source',
                        help='the package to evaluate')
    parser.add_argument('--no-cache', dest='do_not_cache', action='store_const',
                        const=True, default=False,
                        help='Do not cache excuses')
    parser.add_argument('--refresh', action='store_const',
                        const=True, default=False,
                        help='Force refresh of cached excuses')
    parser.add_argument('--debug', action='store_const',
                        const=True, default=False,
                        help='Show debugging information for this tool.')

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG, format="%(message)s")
    else:
        logging.basicConfig(level=logging.INFO, format="%(message)s")

    lp_cachedir = os.path.expanduser(os.path.join("~", ".launchpadlib/cache"))
    lp = Launchpad.login_anonymously(
        'what-next', 'production', lp_cachedir, version='devel')

    refresh_due = False
    xdg_cache = os.getenv('XDG_CACHE_HOME', '~/.cache')
    excuses_path = os.path.expanduser(os.path.join(xdg_cache, 'excuses.yaml'))
    if args.do_not_cache:
        fp = tempfile.NamedTemporaryFile()
    else:
        try:
            fp = open(excuses_path, 'r')
        except FileNotFoundError:
            refresh_due = True
            pass
        finally:
            fp = open(excuses_path, 'a+')

        file_state = os.stat(excuses_path)
        mtime = file_state.st_mtime
        now = time.time()
        if (now - mtime) > MAX_CACHE_AGE:
            refresh_due = True

    with fp:
        if args.refresh or refresh_due:
            url_opener = FancyURLopener()
            excuses_url = ARCHIVE_PAGES + 'proposed-migration/update_excuses.yaml'
            excuses_data = url_opener.retrieve(excuses_url,
                                               fp.name,
                                               report_download)
        fp.seek(0)

        # Use the C implementation of the SafeLoader, it's noticeably faster, and
        # here we're dealing with large input files.
        excuses = yaml.load(fp, Loader=yaml.CSafeLoader)

        if args.source is None:
            print("No source package name was provided. The following packages are "
                  "blocked in proposed:\n")
            args.source = choose_blocked_source(excuses)

        find_excuses(lp, args.source, 0, seen)
